#include <kqml/kqmlconf.h>

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <malloc.h>

#include <kqml.h>
#include <kqml/kqmllib.h>

/*
 * Exported Functions:
 * kqml_build_msgv
 * kqml_build_msg
 * kqml_free_msg
 *
 * kqml_get_field_count
 * kqml_get_performative
 * kqml_get_field
 * kqml_get_ith_field
 *
 * kqml_put_performative
 * _kqml_put_performative_nd
 * kqml_put_field
 * kqml_put_embedded_msg
 * _kqml_put_field_nd
 *
 * _kqml_get_reply_with_field
 *
 */

/*
 * kqml_build_msgv
 *
 * Builds a kqml_message by using the arguments. The arguments are 
 * provided in execv style i.e the field and value arrays have
 * NULL as the last value.
 *
 *	Returns pointer to the newly allocated message on success and
 *	NULL on failure. 
 */
kqml_message * 
kqml_build_msgv (char *perf, char *content, char *rec_agent,
	char *reply_with, char *field[], char *value[])
{
	kqml_message *omsg;
	int i;
 
	if (! perf || perf[0] == '\0')
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("kqml_build_msgv: called without a performative.");
		return NULL;
	}
  
	/* allocate a kqml message */
	if ((omsg = (kqml_message *) malloc (sizeof (kqml_message))) == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("kqml_build_msgv: create msg failed.");
		return NULL;
	}

	omsg->connection = 0;
	omsg->parameter_count = 0;
	omsg->performative = NULL;
	omsg->ppv = NULL;

	/* fill in the fields */
	if (kqml_put_performative (omsg, perf) == -1)
	{
		kqml_free_msg (omsg);
		return NULL;
	}

	if (content)
	{
		if (kqml_put_field (":content", omsg, content, -1) <= -1)
		{
			kqml_free_msg (omsg);
			return NULL;
		}
	}
  
	if (rec_agent)
	{
		if (kqml_put_field (":receiver", omsg, rec_agent, -1) <= -1)
		{
			kqml_free_msg (omsg);
			return NULL;
		}
	}

	if (reply_with)
	{
		if (kqml_put_field (":reply-with", omsg, reply_with, -1) <= -1)
		{
			kqml_free_msg (omsg);
			return NULL;
		}
	}

	if (field)
	{
		if (value == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("kqml_build_msgv: emtpy value array.");
			kqml_free_msg (omsg);
			return NULL;
		}

		i = -1;
		while (i++, field[i] != NULL)
		{
			if (field[i][0] == '\0')
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS) 
					("kqml_build_msgv: empty field name.");
				kqml_free_msg (omsg);
				return NULL;
			}
	
			if (value[i] == NULL)
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS) ("kqml_build_msgv: empty value.");
				kqml_free_msg (omsg);
				return NULL;
			}
			if (kqml_put_field (field[i], omsg, value[i], -1) <= -1)
			{
				kqml_free_msg (omsg);
				return NULL;
			}
		}
	}

	/* return it*/
	return omsg;
}

/*
 * kqml_build_msg
 *
 * Similar to kqml_build_msgv. But the arguments are provided as in
 * printf i.e variable arguemnts. A NULL terminates the argument list
 *
 * Returns a pointer to the newly allocated message on success and 
 * NULL on failure
 */
kqml_message * 
kqml_build_msg (char *perf, char *content, char *rec_agent, 
	char *reply_with, ...)
{
	kqml_message *omsg;
	int i, fieldCount;
	char *field, **fieldArray, *value, **valueArray;
	char endOfArg;
	va_list fp;
 
	fieldCount = 0;
	endOfArg = FALSE;
	va_start (fp, reply_with);
	do
	{
		field = va_arg (fp, char *);
		if (field != NULL)
		{
			fieldCount++;
			value = va_arg (fp, char *);
			if (value == NULL)
			{
				_kqml_dbg_msg (KQML_DBG_VERBOSE) 
					("kqml_build_msg: field %s has NULL value. Ignored.", field);
				endOfArg = TRUE;
			}
		}
		else
		{
			endOfArg = TRUE;
		}
	}
	while (! endOfArg);

	va_end (fp);

	fieldArray = malloc (sizeof (char *) * (fieldCount+1));
	if (fieldArray == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("kqml_build_msg: Could not set field array\n");
		return NULL;
	}

	valueArray = malloc (sizeof (char *) * (fieldCount+1));
	if (valueArray == NULL)
	{
		free (fieldArray);
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("kqml_build_msg: Could not set value array\n");
		return NULL;
	}

	i = 0;
	va_start (fp, reply_with);
	for (i = 0; i < fieldCount; i++)
	{
		fieldArray[i] = va_arg (fp, char *);
		valueArray[i] = va_arg (fp, char *);
	}
	fieldArray[i] = NULL;
	valueArray[i] = NULL;

	omsg = kqml_build_msgv (perf, content, rec_agent, reply_with,
		fieldArray, valueArray);

	free (fieldArray);
	free (valueArray);

	/* return it*/
	return omsg;
}

/*
 * kqml_free_msg
 *
 * frees the memory occupied the kqml message
 */
void kqml_free_msg (kqml_message *msg)
{
	kexpression *exp;

	if (msg == NULL)
	{
		return;
	}

	if (msg->performative != NULL)
	{
		free (msg->performative);
	}

	exp = msg->ppv;

	while (exp)
	{
		if (exp->value != NULL)
		{
			free (exp->value);
		}
		if (exp->key != NULL)
		{
			free (exp->key);
		}

		exp = exp->next;
	}

	free (msg);

	return;
}


/*
 * kqml_get_performative
 *
 * Returns the performative associated with the message
 *
 */
char *
kqml_get_performative (kqml_message *message)
{
	return message->performative;
}

/*
 * kqml_get_field
 *
 * Given a character string which is a kqml keyword and a parsed
 * kqml_message, this function returns the value field associated with
 * that keyword in this message.
 *
 * Returns NULL if there is no such field in this message.
 *
 */
char *
kqml_get_field (char *field_name, kqml_message *message, int *value_len,
	kqml_expression_type *value_type)
{
	kexpression *exp = message->ppv;

	while (exp)
	{
		if (kqml_stricmp (exp->key, field_name) == 0)
		{
			if (value_len != NULL)
			{
				*value_len = exp->value_len;
			}
			if (value_type != NULL)
			{
				*value_type = exp->value_type;
			}
			return (exp->value);
		}
		else
		{
			exp = exp->next;
		}
	}

	return (NULL);
}

/*
 * kqml_get_ith_field
 *
 * Similar to kqml_get_field, but steps through the fields one by one.
 * The caller should invoke it first with field_no set to 0 < #fields
 * If the field_no is valid,
 * it sets field_name and value to the current field.
 *
 * Returns NULL if there are no more fields else the value of the current
 * field
 */
char *
kqml_get_ith_field (int field_no, kqml_message *message, char **field_name,
	char **value, int *value_len, kqml_expression_type *value_type)
{
	int i;
	kexpression *exp;

	if (field_no < 0 || field_no >= message->parameter_count)
		return NULL;

	exp = message->ppv;

	for (i = 0; i < field_no; i++)
	{
		if (exp == NULL)
		{
			return NULL;
		}
		else
		{
			exp = exp->next;
		}
	}

	if (exp == NULL)
	{
		return NULL;
	}

	if (field_name)
	{
		*field_name = exp->key;
	}
	if (value)
	{
		*value = exp->value;
	}
	if (value_len)
	{
		*value_len = exp->value_len;
	}
	if (value_type)
	{
		*value_type = exp->value_type;
	}

	return exp->value;
}

int kqml_get_field_count (kqml_message *msg)
{
	return msg->parameter_count;
}
/*
 * kqml_put_performative
 *
 * Sets the performative field of the message with a copy of the value
 * provided.
 * Returns -1 on error, 1 on replacing an existing performative value,
 * 2 if setting it for the first time
 */
int
kqml_put_performative (kqml_message *msg, char *performative)
{
	char *performative_copy;
	int ret_value;

	performative_copy = strdup (performative);
	if (performative_copy == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("kqml_put_performative: copy performative failed.");
		return -1;
	}

	ret_value = _kqml_put_performative_nd (msg, performative_copy);
	if (ret_value == -1)
	{
		free (performative_copy);
	}

	return ret_value;
}

/*
 * _kqml_put_performative_nd
 *
 * Like kqml_put_performative, but *DOES NOT* make a copy of the performative
 * The value passed *should* have been dynamically allocated, as subsequent
 * calls to kqml_put_performative or _kqml_put_performative_nd will free it.
 *
 * Returns -1 on error, 1 on replacing an existing performative value,
 * 2 if setting it for the first time
 */
int
_kqml_put_performative_nd (kqml_message *msg, char *performative)
{
	if (msg->performative != NULL)
	{
		free (msg->performative);
		msg->performative = performative;
		return 1;
	}
	else
	{
		msg->performative = performative;
		return 2;
	}
}

/*
 * kqml_put_field
 *
 * Given a character string which is a kqml keyword and a parsed
 * kqml_message, this function replaces the current value of that field
 * with the value passed in.  If no keyword like this yet exists, then
 * it is created.
 *
 * Returns 1 if an already existing field has been modified, and returns 2
 * if a new field has been added.
 *
 */
int
kqml_put_field (char *field_name, kqml_message *message, char *value, 
	int value_len)
{
	char *value_copy;
	int ret_value;

	if (value != NULL)
	{
		if (value_len < 0)
		{
			value_len = strlen (value);
		}

		/* The extra +1 is for the null value */
		value_copy = malloc (value_len + 1);
		if (value_copy == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) ("kqml_put_field: copy value failed.");
			return -1;
		}

		memcpy (value_copy, value, value_len);
		value_copy[value_len] = '\0';
	}
	else
	{
		value_len = 0;
		value_copy = NULL;
	}

	ret_value = _kqml_put_field_nd (field_name, message, value_copy, value_len, 
		KQML_TYPE_UNKNOWN);

	if (ret_value == -1)
	{
		if (value_copy)
		{
			free (value_copy);
		}
	}

	return ret_value;
}

int
kqml_put_embedded_msg (char *field_name, kqml_message *message, 
	kqml_message *emsg)
{
	int value_len, ret_value;
	char *value_copy;

	value_copy = kqml_sprint_msg (NULL, &value_len, emsg);

	if (value_copy == NULL)
	{
		return -1;
	}

	ret_value = _kqml_put_field_nd (field_name, message, value_copy, 
		value_len-1, KQML_TYPE_SEXPR);

	if (ret_value == -1)
	{
		free (value_copy);
	}

	return ret_value;
}

/*
 * _kqml_put_field_nd
 * Like kqml_put_field, but does NOT make a copy of the value of the
 * field.
 * the value passed *should* have been dynamically allocated, as it will
 * freed by a subsequent call to kqml_put_field or _kqml_put_field_nd 
 * to replace the value.
 */
int
_kqml_put_field_nd (char *field_name, kqml_message *message, char *value,
	int value_len, kqml_expression_type value_type)
{
	kexpression *exp, *prev;

	if (value_len < 0)
	{
		value_len = strlen (value);
	}

	exp = message->ppv;
	prev = NULL;
	while (exp)
	{
		if (kqml_stricmp (exp->key, field_name) == 0)
		{
			free (exp->value);

			if (value)
			{
				exp->value = value;
				exp->value_len = value_len;
				exp->value_type =  value_type;
			}
			else
			{
				free (exp->key);
				if (message->ppv == exp)
				{
					message->ppv = exp->next;
				}
				else
				{
					prev->next = exp->next;
				}
				free (exp);

				message->parameter_count--;
			}
			return 2;
		}
		else
		{
			prev = exp;
			exp = exp->next;
		}
	}

	if (value == NULL)
	{
		return -1;
	}
	
	exp = (kexpression *) malloc (sizeof (kexpression));
	if (exp == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("kqml_put_field: Could not allocate space for new expression.");
		return -1;
	}

	exp->key = strdup (field_name);
	if (exp->key == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("kqml_put_field: copy field failed.");
		free (exp);
		return -1;
	}

	message->parameter_count++;
	if (message->ppv == NULL)
	{
		message->ppv = exp;
	}
	else
	{
		prev->next = exp;
	}

	exp->next = NULL;
	exp->value = value;
	exp->value_len = value_len;
	exp->value_type =  value_type;

	return 1;
}


/*
 * _kqml_reply-with_field
 *
 * Return the reply-with field of the messsage, but return
 * NULL if the value of the field is nil or NIL.
 *
 */
char *
_kqml_get_reply_with_field (kqml_message *message)
{
	char *value = kqml_get_field (":reply-with", message, NULL, NULL);
  
	if (! value) 
		return NULL;

	if (kqml_stricmp (value, "nil") == 0)
		return NULL;

	return value;
}
