/*
 *	(c) Copyright 1993 Paramax Systems Corp.
 *
 *	This material my be reproduced by or for the U.S.
 *	Government pursuant to the copyright license under
 *	the clause at DFARS 252.227-7013   (1991).
 */

#include <kqml/kqmlconf.h>

#include <stdio.h>
#include <malloc.h>
#include <memory.h>
#include <string.h>

#include <kqml.h>
#include <kqml/kqmllib.h>

/*
 * This function is used to print a kqml message to a stream
 */
void
kqml_print_msg (FILE *stream, kqml_message *kmsg)
{
	int i = 0, field_count, value_len;
	char *field, *value;
	kqml_expression_type value_type;

	if (kmsg)
	{
		/* print the open paren and performative */
		field = kqml_get_performative (kmsg);
		if (field)
		{
			fprintf (stream, "%s\n", field);
		}

		/* print all keyword/value pairs */
		field_count = kqml_get_field_count (kmsg);
		for (i = 0; i < field_count; i++)
		{
			kqml_get_ith_field (i, kmsg, &field, &value, &value_len, &value_type);
			if (field && value)
			{
				/* write the keyword */
				fprintf (stream, "   %s ", field);
				fwrite (value, 1, value_len, stream);
				fprintf (stream, " %d %s\n", value_len,
					(value_type == KQML_TYPE_WORD ? "kqml_word" :
					(value_type == KQML_TYPE_STRING ? "kqml_string" :
					(value_type == KQML_TYPE_BYTE ? "kqml_byte" :
					(value_type == KQML_TYPE_QUOTED ? "kqml_quoted" :
					(value_type == KQML_TYPE_COMMA ? "kqml_comma" :
					(value_type == KQML_TYPE_SEXPR ? "kqml_nested" :
						"kqml_unknown")))))));
			}
		}

		/* print the close paren */
		fprintf(stream, "\n");
		fflush(stream);
	}
}

void
kqml_print_kmsg_list (FILE *stream, KQML_KMSG_LIST *kmsg_list)
{
	KQML_KMSG_LIST *temp;

	temp = kmsg_list;
	if (temp)
	{
		do
		{
			kqml_print_msg (stream, temp->kmsg);
			temp = temp->next;
		}
		while (temp != kmsg_list);
	}
}

/* This structure is a duplicate of current (Ver 1.96) kexpression. 
	I could have used the kexpression structure directly.
	But did not use it for the following reasons:
		Did not want this file to depend on the kqmlmsg.h
		Later, when kexpression is converted to a linked list, it
		will have extra fields which we do not need
*/
typedef struct {
	char *field_name;
	char *value;
	int value_len;
	kqml_expression_type value_type;
	} sprint_struct;

char *
kqml_sprint_msg (char *buffer, int *size, kqml_message *kmsg)
{
	char *perf, *temp;
	int i, len, field_count, actual_field_count;
	sprint_struct *exp_list;

	if (kmsg == NULL)
	{
		return NULL;
	}

	field_count = kqml_get_field_count (kmsg);
	exp_list = malloc (sizeof (sprint_struct) * field_count);

	if (exp_list == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("kqml_sprint_msg: Internal buffer error");
		if (size)
			*size = 0;
		return NULL;
	}

	len = 0;
	actual_field_count = 0;
	for (i = 0; i < field_count; i++)
	{
		kqml_get_ith_field (i, kmsg, &exp_list[actual_field_count].field_name, 
			&exp_list[actual_field_count].value, 
			&exp_list[actual_field_count].value_len, 
			&exp_list[actual_field_count].value_type);

		/* If the type is not known, find it */
		if (exp_list[actual_field_count].value_type == KQML_TYPE_UNKNOWN)
		{
			exp_list[actual_field_count].value_type = 
				_kqml_check_type (exp_list[actual_field_count].value, 
					exp_list[actual_field_count].value_len, 
					&exp_list[actual_field_count].value, 
					&exp_list[actual_field_count].value_len);
		}

		if (exp_list[actual_field_count].value_len > 0)
		{
			len += strlen (exp_list[actual_field_count].field_name) + 
				exp_list[actual_field_count].value_len;

			/* In the case of string and byte string, 
				if the "" or # is not there, we have to add it */
			if (exp_list[actual_field_count].value_type == KQML_TYPE_STRING)
			{
				if (exp_list[actual_field_count].value[0] != '"')
					len += 2;
			}
			else if (exp_list[actual_field_count].value_type == KQML_TYPE_BYTE)
			{
				if (exp_list[actual_field_count].value[0] != '#')
					len += 2 + 
						kqml_digit_count (exp_list[actual_field_count].value_len);
			}
			actual_field_count++;
		}
	}

	perf = kqml_get_performative (kmsg);
	len += strlen (perf) + (actual_field_count * 2) + 3;

	if (buffer == NULL)
	{
		if (size)
		{
			*size = len;
		}
		buffer = malloc (len);
		if (buffer == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) ("kqml_sprint_msg: not enough memory");
			free (exp_list);
			return NULL;
		}
	}
	else
	{
		/* buffer should be of length len */
		if (! size)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("kqml_sprint_msg: length of buffer not given");
			free (exp_list);
			return NULL;
		}
		else if (*size != len)
		{
			_kqml_dbg_msg (KQML_DBG_VERBOSE) 
				("kqml_sprint_msg: buffer not big enough");
			*size = len;
			free (exp_list);
			return NULL;
		}
		*size = len;
	}

	/* I dont check for memory overflow and memccpy's return
		value. There should not be any overflows 
		I also always use len for the same reason
	*/

	temp = buffer;
	*temp++ = '(';
	temp = memccpy(temp, perf, 0, len);
	temp--;
	for (i = 0; i < actual_field_count; i++)
	{
		*temp++ = ' ';
		temp = memccpy(temp, exp_list[i].field_name, 0, len);
		temp--;

		*temp++ = ' ';
		if (exp_list[i].value_type == KQML_TYPE_STRING)
		{
			if (exp_list[i].value[0] != '"')
			{
				*temp++ = '"';
			}
		}
		else if (exp_list[i].value_type == KQML_TYPE_BYTE)
		{
			if (exp_list[i].value[0] != '#')
			{
				*temp++ = '#';
				temp += kqml_digit_to_str (exp_list[i].value_len, temp);
				*temp++ = '"';
			}
		}
				
		memcpy(temp, exp_list[i].value, exp_list[i].value_len);
		temp += exp_list[i].value_len;

		if (exp_list[i].value_type == KQML_TYPE_STRING)
		{
			if (exp_list[i].value[0] != '"')
			{
				*temp++ = '"';
			}
		}
	}

	*temp++ = ')';
	*temp = '\0';

	free (exp_list);
	return buffer;
}
