/*
 * Functions for communicating kqml message 
 * structures between the router and the KRIL.
 *
 * _kqml_init_xdr
 * _kqml_fetch_xdr_msg
 * _kqml_send_xdr_msg
 * _kqml_msgs_in_queue
 */

#include <kqml/kqmlconf.h>

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <malloc.h>
#include <memory.h>

#include <rpc/rpc.h>

#if defined __KQML_MACH_SUN__
extern char *sys_errlist[];
#endif

#include <kqml.h>
#include <kqml/kqmllib.h>

static XDR send_xdr, receive_xdr;
static FILE *send_fp, *receive_fp;

static int rread (void *xdr_handle, void *xdr_buffer, u_int size);
static int rwrite (void *xdr_handle, void *xdr_buffer, u_int size);


/* Note:
	rread and rwrite should *NEVER* be passed with size > MAX_INT
	as this will interfere with error checking

	Cannot use streams for reading or writing. Streams buffer and
	XDR buffer clashes
*/
static int rread (void *xdr_fp, void *xdr_buffer, u_int size)
{
	int fd;
	int read_size;
	char *buffer;

	fd = fileno ((FILE *) xdr_fp);
	read_size = 0;
	buffer = xdr_buffer;

	while (size)
	{
		read_size = read (fd, buffer, size);
		if (read_size == -1) 
		{
			if (errno != EINTR)
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS) 
					("xdr_read: %s.", sys_errlist[errno]);
				return -1;
			}
		}
		else
		{
			if (read_size == 0)
			{
				return -1;
			}
			else
				break;
		}
	}

	return read_size;
}

static int rwrite (void *xdr_fp, void *xdr_buffer, u_int size)
{
	int fd;
	int write_size, size_to_write;
	char *buffer;

	fd = fileno ((FILE *)xdr_fp);
	size_to_write= size;
	buffer = xdr_buffer;

	while (size_to_write)
	{
		write_size = write (fd, buffer, size_to_write);
		if (write_size == -1)
		{
			if (errno != EINTR)
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS) 
					("xdr_write: %s.", sys_errlist[errno]);
				return -1;
			}
		}
		else
		{
			buffer += write_size;
			size_to_write -= write_size;
		}
	}

	return size;
}

int 
_kqml_init_xdr (FILE *r_fp, FILE *s_fp)
{
	receive_fp = r_fp;
	send_fp = s_fp;

	send_xdr.x_op = XDR_ENCODE;
	send_xdr.x_ops = NULL;
	xdrrec_create (&send_xdr, 0, 0, s_fp, rread, rwrite);
	if (send_xdr.x_ops == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("send xdr creation failed. %s.", 
			sys_errlist[errno]);
		return FALSE;
	}

	receive_xdr.x_op = XDR_DECODE;
	receive_xdr.x_ops = NULL;
	xdrrec_create (&receive_xdr, 0, 0, r_fp, rread, rwrite);
	if (receive_xdr.x_ops == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("receive xdr creation failed. %s.",
			sys_errlist[errno]);
		return FALSE;
	}

	return TRUE;
}

kqml_message *
_kqml_fetch_xdr_msg ()
{
	kqml_message *kmsg;
  
	kmsg = calloc (sizeof (kqml_message), 1);
	if (!kmsg) 
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("_kqml_fetch_xdr_msg: Unable to allocate space.\n");
		return NULL;
	}
  
	if (!xdrrec_skiprecord (&receive_xdr))
	{
		kqml_free_msg (kmsg);
		kmsg = NULL;
	}
 	else if (!xdr_kqml_message (&receive_xdr, kmsg))
	{
		kqml_free_msg (kmsg);
		kmsg = NULL;
	}
  
	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("%s is receiving via xdr:\n", 
			(_kqml_im_the_router ? "Router" : "Agent "));
	_kqml_dbg_kmsg (KQML_DBG_VERBOSE) (kmsg);

	return kmsg;
}

/*
 * Send a kqml message to a specified XDR.
 *
 */
int
_kqml_send_xdr_msg (kqml_message *kmsg)
{
	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("%s is sending via xdr:\n", 
			(_kqml_im_the_router ? "Router" : "Agent "));
	_kqml_dbg_kmsg (KQML_DBG_VERBOSE) (kmsg);

	if (!xdr_kqml_message (&send_xdr, kmsg))
		return FALSE;

	return xdrrec_endofrecord (&send_xdr, 1);
}

int
_kqml_msgs_in_queue ()
{
	return ! xdrrec_eof (&receive_xdr);
}

int
_kqml_wait_for_msg (int timeout)
{
	return _kqml_await (fileno (receive_fp), timeout);
}
